# -*- coding: utf-8 -*-
# Original sources Copyright (c) 2006 by Tim Sutton#
# ported to Python by Martin Dobias#
# licensed under the terms of GNU GPL 2#
# Adaptado por German Carrillo 2009#

from PyQt4.QtCore import Qt, QFileInfo, SIGNAL, QPoint, QTimer, QSettings, QVariant
from PyQt4.QtGui import ( QMainWindow, QColor, QVBoxLayout, QDockWidget, QLabel,
                          QFrame, QActionGroup, QFileDialog, QMessageBox )

from qgis.core import QGis, QgsCoordinateReferenceSystem, QgsMapLayerRegistry, QgsPoint, QgsRenderer
from qgis.gui import QgsMapCanvas, QgsMapOverviewCanvas, QgsMapToolPan, QgsMapToolZoom, QgsMapTip, QgsQuickPrint

from formas.visor_ui import Ui_Visor
from utilidades import Utilidades
from cargue import Cargue
from leyenda import Legend
from map_tool_identificar import IdentificarMapTool
from map_tool_medir import MedirDistanciaMapTool

from dialogos.plugins_manager import Plugins

class Visor( QMainWindow, Ui_Visor ):
    """ Clase Visor """
    def __init__( self ):
        # Inicializa la interfaz del visor
        QMainWindow.__init__( self )
        self.setupUi( self )

        self.crearCanvas()
        self.crearLeyenda()
        self.crearMapOverview()
        self.agregarControlesStatusBar()
        self.crearToolsParaElMapa()

        # Instancias a utilizar
        self.cargue = Cargue( self, self.canvas )
        self.utils = Utilidades.obtenerInstancia()

        self.definirSignalsSlots()
        self.agruparTools()
        self.crearMapTips()

        self.plugins = Plugins( self, self.canvas )
        self.cargarSettings()

    def acercar( self ):
        """ Acercar """
        self.canvas.setMapTool( self.toolAcercar )

    def alejar( self ):
        """ Alejar """
        self.canvas.setMapTool( self.toolAlejar )

    def panear( self ):
        """ Panear """
        self.canvas.setMapTool( self.toolPanear )

    def identificar( self ):
        """ Identificar """
        self.canvas.setMapTool( self.toolIdentificar )

    def medir( self ):
        """ Medir sobre el mapa """
        self.canvas.setMapTool( self.toolMedir )

    def vistaCompleta( self ):
        """ Vista total """
        self.canvas.zoomToFullExtent()

    def zoomACapaActiva( self ):
        """ Zoom a la capa activa """
        if self.activeLayer():
            extent = self.activeLayer().layer().extent()
            extent.scale( 1.05 )
            self.canvas.setExtent( extent )
            self.canvas.refresh()

    def zoomASeleccion( self ):
        """ Zoom a la seleccin """
        if self.activeLayer():
            self.canvas.setCurrentLayer( self.activeLayer().layer() )
            self.canvas.zoomToSelected()

    def vistaAnterior( self ):
        """ Vista anterior """
        self.canvas.zoomToPreviousExtent()

    def vistaSiguiente( self ):
        """ Vista siguiente """
        self.canvas.zoomToNextExtent()

    def mostrarMapTip( self ):
        """ SLOT. Muestra los MapTips en el mapa """
        self.timerMapTips.stop() # Parar el Timer

        if self.canvas.underMouse(): # Solo si el mouse est sobre el canvas
            if self.actionMostrarMapTips.isChecked() and self.actionMostrarMapTips.isEnabled():
                puntoQgs = self.lastMapPosition
                puntoQt = self.canvas.mouseLastXY()
                self.mapTip.showMapTip( self.activeLayer().layer(), puntoQgs, puntoQt,
                    self.canvas )

    def exportarMapaAImagen( self ):
        """ Guardar el mapa como imagen (png, jpg o bmp)"""
        bRuta = 0
        settings = QSettings()

        while ( not bRuta ): # Se valida la ruta ingresada
            ultimaRutaExportarAImagen = settings.value( 'Rutas/ExportarAImagen', QVariant('.') ).toString()
            rutaImagen = QFileDialog.getSaveFileName( self,
                "Guardar mapa como imagen.", ultimaRutaExportarAImagen,
                "Imgenes (*.png *.jpg *.bmp)" )

            if rutaImagen:
                imagenInfo = QFileInfo( rutaImagen )
                settings.setValue( 'Rutas/ExportarAImagen', QVariant( imagenInfo.absolutePath() ) )

                if ( ( not rutaImagen.toUpper().endsWith( '.PNG' ) ) \
                    and ( not rutaImagen.toUpper().endsWith( '.JPG' ) ) \
                    and ( not rutaImagen.toUpper().endsWith( '.BMP' ) ) ):
                    bRuta = 0
                    self.utils.mostrarMensaje( self, 'Exportar mapa a imagen',
                        'Debes escribir el nombre de la imagen y su extensin.' \
                        + '\nLas extensiones posibles son png, jpg y bmp.', QMessageBox.Warning )
                else: # Exportar
                    bRuta = 1
                    self.canvas.saveAsImage( rutaImagen, self.canvas.canvasPixmap(),
                        str( rutaImagen[ -3: ] ).upper() )
                    self.utils.mostrarMensaje( self, 'Exportar mapa a imagen',
                        'El mapa ha sido exportado exitosamente.',
                        QMessageBox.Information )
            else: # Click en cancelar
                bRuta = 1

    def mostrarEscala( self, escala ):
        """ SLOT. Muestra la escala en la Barra de Estado """
        self.lblEscala.setText( "Escala 1:" + self.utils.formatearNumero( escala ) )

    def mostrarXY( self, p ):
        """ SLOT. Muestra las coordenadas del cursor en la Barra de estado """
        if self.canvas.mapUnits() == 2: # Grados
            self.lblXY.setText( self.utils.formatearAGrados( p.x() ) + " | " \
                + self.utils.formatearAGrados( p.y() ) )
        else: # Unidad lineal
            self.lblXY.setText( self.utils.formatearNumero( p.x() ) + " m. | " \
                + self.utils.formatearNumero( p.y() ) + " m." )

    def mapTipXYChanged( self, p ):
        """ SLOT. Inicializa el Timer para mostrar MapTips en el mapa """
        if self.canvas.underMouse(): # Solo si el mouse est sobre el canvas
            if self.actionMostrarMapTips.isChecked() and self.actionMostrarMapTips.isEnabled():
                self.lastMapPosition = QgsPoint( p.x(), p.y() )
                #self.timerMapTips.stop() # Parar el Timer
                self.mapTip.clear( self.canvas )
                self.timerMapTips.start( 700 )

    def definirSRSMapa( self ):
        """ Define el SRS para las capas o las reproyecta """
        if self.cargue.bSRSDefinido: # Reproyectar
            self.canvas.mapRenderer().setProjectionsEnabled( True )
        else: # Asignar proyeccin
            self.canvas.mapRenderer().setProjectionsEnabled( False )

        codEpsg = self.utils.definirSRS( self, "mapa" ) # Abrir dilogo SRS

        if codEpsg:
            srs = QgsCoordinateReferenceSystem( codEpsg,
                QgsCoordinateReferenceSystem.EpsgCrsId )
            self.canvas.mapRenderer().setDestinationSrs( srs )

            if not self.cargue.bSRSDefinido: # Asignar SRS a las capas
                for i in range( self.canvas.layerCount() ):
                    self.canvas.layer( i ).setCrs( srs )

            # Definir unidades del mapa
            if srs.geographicFlag():
                self.canvas.setMapUnits( QGis.Degrees )
            else:
                self.canvas.setMapUnits( QGis.Meters )

            # Emitir SIGNAL con las nuevas unidades como argumento
            self.emitirSignalUnidadesMapaCambiadas()

            #Actualizar escala y redefinir el extent
            self.canvas.updateScale()
            self.canvas.zoomToFullExtent()
            self.cargue.bSRSDefinido = True # Se puede reproyectar
        else:
            self.utils.mostrarMensaje( self, "Sistema de Referencia",
                "No has definido un Sistema de \nReferencia para el mapa.",
                QMessageBox.Warning )

    def emitirSignalUnidadesMapaCambiadas( self ):
        """ Emite SIGNAL con las nuevas unidades como argumento """
        self.emit( SIGNAL( "unidadesMapaCambiadas" ), self.canvas.mapUnits() )

    def abrirGeoTux( self ):
        res = self.utils.mostrarMensajeConfirmacion( self, "Visitar GeoTux",
            "A continuacin se abrir la pgina de GeoTux \nen tu navegador " \
            + "web predeterminado. \n\nEsts de acuerdo?" )
        if res == "si":
            import webbrowser
            webbrowser.open( "http://geotux.tuxfamily.org" )

    def acercaDe( self )  :
        QMessageBox.about( self, "Acerca de Visor GeoTux",
            "              Visor de datos geogrficos.\n                    " \
            + "PyQt4 y PyQGIS.\n\nGermn Carrillo " \
            + "<geotux_tuxman@linuxmail.org>\n               " \
            + "http://geotux.tuxfamily.org\n                  " \
            + "Licencia: GNU/GPL v.3" )

    def mostrarTodasLasCapas( self ):
        """ Muestra todas las capas en el mapa """
        self.legend.setStatusForAllLayers( True )

    def ocultarTodasLasCapas( self ):
        """ Oculta todas las capas en el mapa """
        self.legend.setStatusForAllLayers( False )

    def removerTodasLasCapas( self ):
        """ Remueve todas las capas del mapa """
        res = self.utils.mostrarMensajeConfirmacion( self,
            "Usos Alternativos", "Remover todas las capas del mapa?" )
        if res == "si":
            QgsMapLayerRegistry.instance().removeAllMapLayers()

    def mostrarTodoEnVistaGeneral( self ):
        """ Muestra todas las capas en la vista general """
        self.legend.allLayersInOverview( True )

    def removerTodoDeVistaGeneral( self ):
        """ Remueve todas las capas de la vista general """
        self.legend.allLayersInOverview( False )

    def crearCanvas( self ):
        """ Crea el map canvas """
        self.canvas = QgsMapCanvas()
        self.canvas.setCanvasColor( QColor( 255,255,255 ) )
        self.canvas.enableAntiAliasing( True )
        self.canvas.useImageToRender( False )
        self.canvas.show()
        QgsRenderer.setSelectionColor( QColor( 85, 255, 255 ) )

        # Se agrega el Map canvas a la ventana principal, dentro del marco
        self.layout = QVBoxLayout( self.frame )
        self.layout.setMargin( 0 )
        self.layout.addWidget( self.canvas )

    def crearLeyenda( self ):
        """ Crea el control Leyenda y lo asocia al canvas """
        self.legend = Legend( self )
        self.legend.setCanvas( self.canvas )
        self.legend.setObjectName( "theMapLegend" )

        self.LegendDock = QDockWidget( "Tabla de Contenido", self )
        self.LegendDock.setObjectName( "legend" )
        self.LegendDock.setAllowedAreas( Qt.LeftDockWidgetArea | Qt.RightDockWidgetArea )
        self.LegendDock.setWidget( self.legend )
        self.addDockWidget( Qt.LeftDockWidgetArea, self.LegendDock )

    def crearMapOverview( self ):
        """ Crea el map canvas OverView """
        self.canvasOV = QgsMapOverviewCanvas( self, self.canvas )
        self.canvasOV.enableAntiAliasing( False ) # False: Mejora el rendimiento
        self.canvasOV.setCursor( Qt.OpenHandCursor )

        self.CanvasOVDock = QDockWidget( "Vista General", self )
        self.CanvasOVDock.setObjectName( "canvasOV" )
        self.CanvasOVDock.setAllowedAreas( Qt.LeftDockWidgetArea | Qt.RightDockWidgetArea )
        self.CanvasOVDock.setMinimumHeight( 120 )
        self.CanvasOVDock.setWidget( self.canvasOV )
        self.addDockWidget( Qt.LeftDockWidgetArea, self.CanvasOVDock )

        self.canvas.enableOverviewMode( self.canvasOV )

    def agregarControlesStatusBar( self ):
        """ Agrega widgets al StatusBar """
        self.lblXY = QLabel()
        self.lblXY.setFrameStyle( QFrame.Box )
        self.lblXY.setMinimumWidth( 170 )
        self.lblXY.setAlignment( Qt.AlignCenter )
        self.statusbar.setSizeGripEnabled( False )
        self.statusbar.addPermanentWidget( self.lblXY, 0 )

        self.lblEscala = QLabel()
        self.lblEscala.setFrameStyle( QFrame.StyledPanel )
        self.lblEscala.setMinimumWidth( 140 )
        self.statusbar.addPermanentWidget( self.lblEscala, 0 )

    def crearToolsParaElMapa( self ):
        """ Crea las herramientas (tools) para el mapa """
        self.toolAcercar = QgsMapToolZoom( self.canvas, False ) # false=Acercar
        self.toolAlejar = QgsMapToolZoom( self.canvas, True ) # true=Alejar
        self.toolPanear = QgsMapToolPan( self.canvas )
        self.toolIdentificar = IdentificarMapTool( self, self.canvas )
        self.toolMedir = MedirDistanciaMapTool( self.canvas, self )

        self.canvas.setMapTool( self.toolPanear ) # Tool activa por defecto
        self.actionPanear.setChecked( True )

    def definirSignalsSlots( self ):
        """ SIGNALS - SLOTS """
        self.connect( self.actionAcercar, SIGNAL( "activated()" ), self.acercar )
        self.connect( self.actionAlejar, SIGNAL( "activated()" ), self.alejar )
        self.connect( self.actionPanear, SIGNAL( "activated()" ), self.panear )
        self.connect( self.actionVistaCompleta, SIGNAL( "activated()" ),
            self.vistaCompleta )
        self.connect( self.actionZoomACapaActiva, SIGNAL( "activated()" ),
            self.zoomACapaActiva )
        self.connect( self.actionZoomASeleccion, SIGNAL( "activated()" ),
            self.zoomASeleccion )
        self.connect( self.actionVistaAnterior, SIGNAL( "activated()" ),
            self.vistaAnterior )
        self.connect( self.actionVistaSiguiente, SIGNAL( "activated()" ),
            self.vistaSiguiente )

        self.connect( self.actionExportarMapaAImagen,
            SIGNAL( "activated()" ), self.exportarMapaAImagen )
        self.connect( self.actionDefinirSRSMapa, SIGNAL( "activated()" ),
            self.definirSRSMapa )
        self.connect( self.actionCargarShapefile, SIGNAL( "activated()" ),
            self.cargue.cargarShapefile )
        self.connect( self.actionCargarPostGIS, SIGNAL( "activated()" ),
            self.cargue.cargarPostGIS )
        self.connect( self.actionCargarRaster, SIGNAL( "activated()" ),
            self.cargue.cargarRaster )
        self.connect( self.actionCargarWMS, SIGNAL( "activated()" ),
            self.cargue.cargarWMS )
        self.connect( self.actionSalir, SIGNAL( "activated()" ), self.salir )

        self.connect( self.canvas, SIGNAL( "scaleChanged(double)" ),
            self.mostrarEscala )
        self.connect( self.canvas, SIGNAL( "xyCoordinates(QgsPoint&)" ),
            self.mostrarXY )

        self.connect( self.actionIdentificar, SIGNAL( "activated()" ),
            self.identificar )

        self.connect( self.actionMostrarTodasLasCapas, SIGNAL( "activated()" ),
            self.mostrarTodasLasCapas )
        self.connect( self.actionOcultarTodasLasCapas, SIGNAL( "activated()" ),
            self.ocultarTodasLasCapas )
        self.connect( self.actionRemoverTodasLasCapas, SIGNAL( "activated()" ),
            self.removerTodasLasCapas )
        self.connect( self.actionMostrarTodoEnVistaGeneral, SIGNAL( "activated()"),
            self.mostrarTodoEnVistaGeneral )
        self.connect( self.actionRemoverTodoDeVistaGeneral, SIGNAL( "activated()"),
            self.removerTodoDeVistaGeneral )

        self.connect( self.actionMedirDistancia, SIGNAL( "activated()" ),
            self.medir )
        self.connect( self.actionVisitaGeoTux, SIGNAL( "activated()" ),
            self.abrirGeoTux)
        self.connect( self.actionAcercaDe, SIGNAL( "activated()" ),
            self.acercaDe )

        self.connect( self.legend, SIGNAL( "activeLayerChanged" ),
            self.habilitarHerramientas )
        self.connect( QgsMapLayerRegistry.instance(), SIGNAL( "removedAll()" ),
            self.legend.removeAll )
        self.connect( QgsMapLayerRegistry.instance(), SIGNAL( "layerWillBeRemoved(QString)" ),
            self.toolIdentificar.capaRemovida )

    def agruparTools( self ):
        """ Agrupa las tools en un grupo para asegurar su activacin exclusiva """
        acciones = QActionGroup( self )
        acciones.addAction( self.actionAcercar )
        acciones.addAction( self.actionAlejar )
        acciones.addAction( self.actionPanear )
        acciones.addAction( self.actionIdentificar )
        acciones.addAction( self.actionMedirDistancia )

    def crearMapTips( self ):
        """ Crea los MapTips en el mapa """
        self.timerMapTips = QTimer( self.canvas )
        self.mapTip = QgsMapTip()
        self.connect( self.canvas, SIGNAL( "xyCoordinates(QgsPoint&)" ),
            self.mapTipXYChanged )
        self.connect( self.timerMapTips, SIGNAL( "timeout()" ),
            self.mostrarMapTip )

    def habilitarHerramientas( self ):
        """ SLOT. Habilita o no herramientas partiendo de condiciones """
        if not self.activeLayer():
            self.actionMedirDistancia.setEnabled( False )
            self.actionDefinirSRSMapa.setEnabled( False )
            self.actionIdentificar.setEnabled( False )
            self.actionMostrarMapTips.setEnabled( False )
            self.actionMostrarTodasLasCapas.setEnabled( False )
            self.actionOcultarTodasLasCapas.setEnabled( False )
            self.actionRemoverTodasLasCapas.setEnabled( False )
            self.actionMostrarTodoEnVistaGeneral.setEnabled( False )
            self.actionRemoverTodoDeVistaGeneral.setEnabled( False )
        else:
            self.actionMedirDistancia.setEnabled( True )
            self.actionDefinirSRSMapa.setEnabled( True )
            self.actionIdentificar.setEnabled( True )
            self.actionMostrarTodasLasCapas.setEnabled( True )
            self.actionOcultarTodasLasCapas.setEnabled( True )
            self.actionRemoverTodasLasCapas.setEnabled( True )
            self.actionMostrarTodoEnVistaGeneral.setEnabled( True )
            self.actionRemoverTodoDeVistaGeneral.setEnabled( True )

            tipoCapa = self.activeLayer().layer().type()
            if tipoCapa == 0: # Vector Layer
                self.actionMostrarMapTips.setEnabled( True )
            elif tipoCapa == 1: # Raster Layer
                self.actionMostrarMapTips.setEnabled( False )

    def activeLayer( self ):
        """ Retorna la capa activa de la leyenda """
        return self.legend.activeLayer()

    def refreshLayerSymbology( self, layer ):
        self.legend.refreshLayerSymbology( layer )

    def cargarSettings( self ):
        """
            Carga las preferencias del usuario con respecto a la ventana principal
             - Tamao y posicin de la ventana
             - Posiciones de Toolbars y DockedWindows
        """
        settings = QSettings()
        #if settings.contains( "Visor/Geometry" ): # OJO
        self.restoreGeometry( settings.value( "Visor/Geometry" ).toByteArray() )
        self.restoreState( settings.value( "Visor/Properties" ).toByteArray() )

    def guardarSettings( self ):
        """
            Carga las preferencias del usuario con respecto a la ventana principal
             - Tamao y posicin de la ventana
             - Posiciones de Toolbars y DockedWindows
        """
        settings = QSettings()
        settings.setValue( "Visor/Geometry", QVariant( self.saveGeometry() ) )
        settings.setValue( "Visor/Properties", QVariant( self.saveState() ) )

    def closeEvent( self, event ):
        """ Manejador del evento Close """
        res = self.utils.mostrarMensajeConfirmacion( self,
            "Visor GeoTux", "Realmente quieres salir?" )
        if res == "si":
            self.guardarSettings()
            self.emit( SIGNAL( 'cerrandoVisor()' ) ) # Emitir SIGNAL de cierre
            event.accept()
        elif res == "no":
            event.ignore()

    def salir( self ):
        self.close()
